﻿/*
 * Fluid Dynamics
 * Copyright (C) 2008 Michael Birken
 * 
 * This file is part of Fluid Dynamics.
 *
 * Fluid Dynamics is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fluid Dynamics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Drawing.Drawing2D;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Threading;

namespace FluidDynamics {

  public partial class FluidForm : Form {

    [StructLayout(LayoutKind.Sequential)]
    public struct PeekMsg {
      public IntPtr hWnd;
      public Message msg;
      public IntPtr wParam;
      public IntPtr lParam;
      public uint time;
      public System.Drawing.Point p;
    }

    [System.Security.SuppressUnmanagedCodeSecurity]
    [DllImport("User32.dll", CharSet = CharSet.Auto)]
    public static extern bool PeekMessage(out PeekMsg msg, IntPtr hWnd,
             uint messageFilterMin, uint messageFilterMax, uint flags);

    private const int IMAGE_SCALE = 8;
    private const int WIDTH = 80;
    private const int HEIGHT = 40;

    private const double DT = 0.2;
    private const int JACOBI_ITERATIONS = 40;
    private const double VISCOSITY = 0.000000000001;
    private const double MOUSE_MASS = 0.01;
    private const double DYE_REMOVAL_RATE = 0.998;

    private const double DIFFUSION_ALPHA = 1.0 / (VISCOSITY * DT);
    private const double DIFFUSION_BETA_RECIPROCAL = 1.0 / (4.0 + DIFFUSION_ALPHA);
    private const double PRESSURE_ALPHA = -1.0;
    private const double PRESSURE_BETA_RECIPROCAL = 1.0 / 4.0;
    private const double MOUSE_MASS_RECIPROCAL = 1.0 / MOUSE_MASS;
    private const int IMAGE_WIDTH = IMAGE_SCALE * WIDTH;
    private const int IMAGE_HEIGHT = IMAGE_SCALE * HEIGHT;    
    private Bitmap bitmap;
    private Graphics graphics;
    private Rectangle rect;
    private uint[] pixels = new uint[WIDTH * HEIGHT];
    private bool mouseDown;
    private int lastMouseX, lastMouseY, deltaMouseX, deltaMouseY;
    private int inkColorIndex;

    private double[, ,] velocities = new double[HEIGHT, WIDTH, 2];
    private double[,] pressures = new double[HEIGHT, WIDTH];
    private double[,,] concentrations = new double[HEIGHT, WIDTH, 3];

    private double[, ,] _velocities = new double[HEIGHT, WIDTH, 2];
    private double[,,] _concentrations = new double[HEIGHT, WIDTH, 3];
    private double[,] _pressures = new double[HEIGHT, WIDTH];
    private double[,] divergences = new double[HEIGHT, WIDTH];

    private void GetVelocity(double x, double y, out double vx, out double vy) {
      int floorX = (int)Math.Floor(x);
      int ceilingX = (int)Math.Ceiling(x);
      int floorY = (int)Math.Floor(y);
      int ceilingY = (int)Math.Ceiling(y);
      double percentX = x - floorX;
      double percentY = y - floorY;
      double upperLeftX, upperLeftY;
      double upperRightX, upperRightY;
      double lowerLeftX, lowerLeftY;
      double lowerRightX, lowerRightY;
      ReadVelocity(floorX, floorY, out upperLeftX, out upperLeftY);
      ReadVelocity(ceilingX, floorY, out upperRightX, out upperRightY);
      ReadVelocity(floorX, ceilingY, out lowerLeftX, out lowerLeftY);
      ReadVelocity(ceilingX, ceilingY, out lowerRightX, out lowerRightY);
      vx = GetValue(percentX, percentY, upperLeftX, upperRightX, lowerLeftX, lowerRightX);
      vy = GetValue(percentX, percentY, upperLeftY, upperRightY, lowerLeftY, lowerRightY);
    }

    private void ReadVelocity(int x, int y, out double vx, out double vy) {
      if (x < 0 || y < 0 || x >= WIDTH || y >= HEIGHT) {
        vx = 0;
        vy = 0;
      } else {
        vx = velocities[y, x, 0];
        vy = velocities[y, x, 1];
      }
    }

    private double GetPressure(int x, int y) {
      if (x < 0) {
        x = 0;
      } else if (x >= WIDTH) {
        x = WIDTH - 1;
      }
      if (y < 0) {
        y = 0;
      } else if (y >= HEIGHT) {
        y = HEIGHT - 1;
      }
      return pressures[y, x];
    }

    private void ComputeDivergences() {
      for (int y = 0; y < HEIGHT; y++) {
        for (int x = 0; x < WIDTH; x++) {
          double v0x, v0y;
          double v1x, v1y;
          double v2x, v2y;
          double v3x, v3y;
          ReadVelocity(x, y - 1, out v0x, out v0y);
          ReadVelocity(x, y + 1, out v1x, out v1y);
          ReadVelocity(x - 1, y, out v2x, out v2y);
          ReadVelocity(x + 1, y, out v3x, out v3y);
          divergences[y, x] = 0.5 * ((v3x - v2x) + (v1y - v0y));
        }
      }
    }

    private double GetConcentration(double x, double y, int colorIndex) {
      int floorX = (int)Math.Floor(x);
      int ceilingX = (int)Math.Ceiling(x);
      int floorY = (int)Math.Floor(y);
      int ceilingY = (int)Math.Ceiling(y);
      return GetValue(x - floorX, y - floorY,
          GetConcentrationInContainer(floorX, floorY, colorIndex),
          GetConcentrationInContainer(ceilingX, floorY, colorIndex),
          GetConcentrationInContainer(floorX, ceilingY, colorIndex),
          GetConcentrationInContainer(ceilingX, ceilingY, colorIndex));
    }

    private double GetConcentrationInContainer(int x, int y, int colorIndex) {
      if (x < 0 || y < 0 || x >= WIDTH || y >= HEIGHT) {
        return 0;
      }
      return concentrations[y, x, colorIndex];
    }

    private double GetValue(double percentX, double percentY, 
        double upperLeft, double upperRight, double lowerLeft, double lowerRight) {
      double upper = (upperRight - upperLeft) * percentX + upperLeft;
      double lower = (lowerRight - lowerLeft) * percentX + lowerLeft;
      return (lower - upper) * percentY + upper;
    }

    public FluidForm() {
      InitializeComponent();
      InitFrame();
    }

    private void FluidForm_Load(object sender, EventArgs e) {
      InitImage();
      Application.Idle += OnApplicationIdle;
    }

    private void OnApplicationIdle(object sender, EventArgs e) {
      PeekMsg msg;
      while (!PeekMessage(out msg, IntPtr.Zero, 0, 0, 0)) {
        UpdateModel();
        RenderFrame();
      }
    }

    private void SwapConcentrations() {
      double[,,] temp = concentrations;
      concentrations = _concentrations;
      _concentrations = temp;
    }

    private void SwapPressures() {
      double[,] temp = pressures;
      pressures = _pressures;
      _pressures = temp;
    }

    private void SwapVelocities() {
      double[, ,] temp = velocities;
      velocities = _velocities;
      _velocities = temp;
    }

    private void AdvectDyeConcentration() {
      for (int y = 0; y < HEIGHT; y++) {
        for (int x = 0; x < WIDTH; x++) {
          for (int colorIndex = 0; colorIndex < 3; colorIndex++) {
            _concentrations[y, x, colorIndex] = DYE_REMOVAL_RATE * Math.Min(255.0, GetConcentration(
                x - velocities[y, x, 0] * DT,
                y - velocities[y, x, 1] * DT,
                colorIndex));
          }
        }
      }

      SwapConcentrations();
    }

    private void Advect() {
      for (int y = 0; y < HEIGHT; y++) {
        for (int x = 0; x < WIDTH; x++) {
          GetVelocity(
            x - velocities[y, x, 0] * DT,
            y - velocities[y, x, 1] * DT,
            out _velocities[y, x, 0],
            out _velocities[y, x, 1]);
        }
      }

      SwapVelocities();
    }

    private void Diffuse() {
      for (int jacobis = 0; jacobis < JACOBI_ITERATIONS; jacobis++) {
        for (int y = 0; y < HEIGHT; y++) {
          for (int x = 0; x < WIDTH; x++) {
            double v0x, v0y;
            double v1x, v1y;
            double v2x, v2y;
            double v3x, v3y;
            double vx, vy;
            ReadVelocity(x, y - 1, out v0x, out v0y);
            ReadVelocity(x, y + 1, out v1x, out v1y);
            ReadVelocity(x - 1, y, out v2x, out v2y);
            ReadVelocity(x + 1, y, out v3x, out v3y);
            ReadVelocity(x, y, out vx, out vy);
            _velocities[y, x, 0] = (v0x + v1x + v2x + v3x + DIFFUSION_ALPHA * vx) * DIFFUSION_BETA_RECIPROCAL;
            _velocities[y, x, 1] = (v0y + v1y + v2y + v3y + DIFFUSION_ALPHA * vy) * DIFFUSION_BETA_RECIPROCAL;
          }
        }
        SwapVelocities();
      }
    }

    private void AddForces() {
      if (mouseDown) {
        int x = lastMouseX / IMAGE_SCALE;
        int y = lastMouseY / IMAGE_SCALE;
        if (x < 0 || y < 0 || x >= IMAGE_HEIGHT || y >= IMAGE_WIDTH) {
          return;
        }
        velocities[y, x, 0] += DT * MOUSE_MASS_RECIPROCAL * deltaMouseX;
        velocities[y, x, 1] += DT * MOUSE_MASS_RECIPROCAL * deltaMouseY;
      }
    }

    private void ComputePressure() {

      ComputeDivergences();

      for (int y = 0; y < HEIGHT; y++) {
        for (int x = 0; x < WIDTH; x++) {
          pressures[y, x] = 0;
        }
      }

      for (int jacobis = 0; jacobis < JACOBI_ITERATIONS; jacobis++) {
        for (int y = 0; y < HEIGHT; y++) {
          for (int x = 0; x < WIDTH; x++) {
            _pressures[y, x] = (GetPressure(x, y - 1)
                + GetPressure(x, y + 1)
                + GetPressure(x - 1, y)
                + GetPressure(x + 1, y)
                + PRESSURE_ALPHA * divergences[y, x]) * PRESSURE_BETA_RECIPROCAL;
          }
        }
        SwapPressures();
      }
    }

    private void SubtractPressureGradient() {
      for (int y = 0; y < HEIGHT; y++) {
        for (int x = 0; x < WIDTH; x++) {
          velocities[y, x, 0] -= 0.5 * (GetPressure(x + 1, y) - GetPressure(x - 1, y));
          velocities[y, x, 1] -= 0.5 * (GetPressure(x, y + 1) - GetPressure(x, y - 1));
        }
      }
    }

    private void InjectDye() {
      if (mouseDown) {
        int y = lastMouseY / IMAGE_SCALE;
        int x = lastMouseX / IMAGE_SCALE;
        for (int i = y - 2; i < y + 4; i++) {
          for (int j = x - 2; j < x + 4; j++) {
            if (i >= 0 && j >= 0 && i < HEIGHT && j < WIDTH) {
              concentrations[i, j, inkColorIndex] = 255.0;
            }
          }
        }
      }
    }

    private void UpdateModel() {
      Advect();
      Diffuse();
      AddForces();
      ComputePressure();
      SubtractPressureGradient();
      InjectDye();
      AdvectDyeConcentration();
    }

    private void RenderFrame() {
      for (int y = 0, pixelsIndex = 0; y < HEIGHT; y++) {
        for (int x = 0; x < WIDTH; x++, pixelsIndex++) {
          uint red = (uint)concentrations[y, x, 0];
          uint green = (uint)concentrations[y, x, 1];
          uint blue = (uint)concentrations[y, x, 2];
          pixels[pixelsIndex] = 0xFF000000 | (red << 16) | (green << 8) | blue;
        }
      }
      graphics.DrawImage(bitmap, 0, 0, IMAGE_WIDTH, IMAGE_HEIGHT);
    }

    private void InitGraphics() {
      graphics = CreateGraphics();
      graphics.CompositingQuality = System.Drawing.Drawing2D.CompositingQuality.HighSpeed;
      graphics.CompositingMode = System.Drawing.Drawing2D.CompositingMode.SourceCopy;
      graphics.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.NearestNeighbor;
    }

    private void InitImage() {
      InitGraphics();
      for (int i = 0; i < pixels.Length; i++) {
        pixels[i] = 0xFF000000;
      }
      GCHandle handle = GCHandle.Alloc(pixels, GCHandleType.Pinned);
      IntPtr pointer = Marshal.UnsafeAddrOfPinnedArrayElement(pixels, 0);
      bitmap = new Bitmap(WIDTH, HEIGHT, 4 * WIDTH, PixelFormat.Format32bppPArgb, pointer);
    }

    private void InitFrame() {
      rect = new Rectangle(0, 0, WIDTH, HEIGHT);
      StartPosition = System.Windows.Forms.FormStartPosition.CenterScreen;
      SetStyle(ControlStyles.UserPaint, true);
      SetStyle(ControlStyles.AllPaintingInWmPaint, true);
      ClientSize = new Size(IMAGE_WIDTH, IMAGE_HEIGHT);
    }

    private void FluidForm_Paint(object sender, PaintEventArgs e) {
      e.Graphics.CompositingQuality = System.Drawing.Drawing2D.CompositingQuality.HighSpeed;
      e.Graphics.CompositingMode = System.Drawing.Drawing2D.CompositingMode.SourceCopy;
      e.Graphics.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.NearestNeighbor;
      e.Graphics.DrawImage(bitmap, 0, 0, IMAGE_WIDTH, IMAGE_HEIGHT);
    }

    private void FluidForm_MouseMove(object sender, MouseEventArgs e) {
      if (mouseDown) {
        int x = e.X;
        int y = e.Y;
        if (x < 0 || y < 0 || x >= IMAGE_WIDTH || y >= IMAGE_HEIGHT) {
          deltaMouseX = 0;
          deltaMouseY = 0;
          return;
        }
        deltaMouseX = x - lastMouseX;
        deltaMouseY = y - lastMouseY;
        lastMouseX = x;
        lastMouseY = y;        
      }
    }

    private void FluidForm_MouseUp(object sender, MouseEventArgs e) {
      mouseDown = false;
      if (++inkColorIndex == 3) {
        inkColorIndex = 0;
      }
    }

    private void FluidForm_MouseDown(object sender, MouseEventArgs e) {
      mouseDown = true;
      lastMouseX = e.X;
      lastMouseY = e.Y;
    }
  }
}
